# -- Imports ------------------------------------------------------------------

from argparse import ArgumentParser
from asyncio import run, sleep
from collections import Counter

from netaddr import EUI

from mytoolit.can import Network
from mytoolit.cmdline import byte_value, mac_address

# -- Function -----------------------------------------------------------------


def parse_arguments():
    """Parse the arguments of the EEPROM checker command line tool

    Returns
    -------

    A simple object storing the MAC address (attribute `mac`) of an STH and an
    byte value that should be stored into the cells of the EEPROM (attribute
    `value`)
    """

    parser = ArgumentParser(
        description="Check the integrity of STH EEPROM content"
    )
    parser.add_argument(
        "mac",
        help="MAC address of STH e.g. 08:6b:d7:01:de:81",
        type=mac_address,
    )
    parser.add_argument(
        "--value",
        help="byte value for EEPROM cells (default: %(default)s)",
        type=byte_value,
        default=10,
    )

    return parser.parse_args()


# -- Class --------------------------------------------------------------------


class EEPROMCheck:
    """Write and check the content of a certain page in EEPROM of an STH"""

    def __init__(self, mac: EUI, value):
        """Initialize the EEPROM check with the given arguments

        Parameters
        ----------

        mac
            The MAC address of an STH

        value:
            The value that the EEPROM checker should write into the EEPROM
        """

        self.mac_address = mac
        self.eeprom_address = 1
        self.eeprom_length = 256
        self.eeprom_value = value

    async def __aenter__(self):
        """Initialize the connection to the STU"""

        self.network = Network()
        await self.network.reset_node("STU 1")
        await sleep(1)  # Wait till reset takes place
        return self

    async def __aexit__(self, exception_type, exception_value, traceback):
        """Disconnect from the STU"""

        await self.network.shutdown()

    async def connect_bluetooth(self):
        """Connect to the STH"""

        await self.network.connect_sensor_device(self.mac_address)
        print(f"Connected to “{await self.network.read_eeprom_name('STH 1')}”")

    async def reset_sth(self):
        """Reset the (connected) STH"""

        await self.network.reset_node("STH 1")
        await sleep(1)  # Wait till reset takes place

    async def write_eeprom(self):
        """Write a byte value into one page of the EEPROM"""

        print(f"Write value “{self.eeprom_value}” into EEPROM cells")
        await self.network.write_eeprom(
            address=1,
            offset=0,
            data=[self.eeprom_value for _ in range(self.eeprom_length)],
            node="STH 1",
        )

    async def read_eeprom(self):
        """Read a page of the EEPROM

        Returns:

        A list of the byte values stored in the EEPROM page
        """

        return await self.network.read_eeprom(
            address=1, offset=0, length=256, node="STH 1"
        )

    async def print_eeprom_incorrect(self):
        """Print a summary of the incorrect values in the EEPROM page"""

        changed = [
            byte
            for byte in await self.read_eeprom()
            if byte != self.eeprom_value
        ]
        incorrect = len(changed) / self.eeprom_length
        counter = Counter(changed)
        summary = ", ".join(
            f"{value} ({times} time{'' if times == 1 else 's'})"
            for value, times in sorted(
                counter.items(), key=lambda item: item[1], reverse=True
            )
        )
        print(f"{incorrect:.2%} incorrect{': ' if summary else ''}{summary}")

    async def print_eeprom(self):
        """Print the values stored in the EEPROM page"""

        page = await self.read_eeprom()
        bytes_per_line = 8
        for byte in range(0, self.eeprom_length - 1, bytes_per_line):
            print(f"{byte:3}: ", end="")
            byte_representation = " ".join(["{:3}"] * bytes_per_line).format(
                *page[byte : byte + bytes_per_line]
            )
            print(byte_representation)


# -- Functions ----------------------------------------------------------------


async def check_eeprom(arguments):
    async with EEPROMCheck(mac=arguments.mac, value=arguments.value) as check:
        await check.connect_bluetooth()
        await check.write_eeprom()
        await check.print_eeprom_incorrect()
        print()
        for _ in range(5):
            await check.reset_sth()
            await check.connect_bluetooth()
            await check.print_eeprom_incorrect()
            print()
        await check.print_eeprom()


# -- Main ---------------------------------------------------------------------


def main():
    run(check_eeprom(parse_arguments()))


if __name__ == "__main__":
    main()



# -- Exports ------------------------------------------------------------------

from .acceleration import convert_raw_to_g, ratio_noise_max
from .voltage import convert_raw_to_supply_voltage
from .constants import ADC_MAX_VALUE
from .storage import Storage
from .units import celsius, g0, volt



import argparse
import multiprocessing
import socket

from argparse import ArgumentDefaultsHelpFormatter
from time import sleep, time
from datetime import datetime
from functools import partial
from logging import getLogger, FileHandler, Formatter
from pathlib import Path
from platform import system
from sys import stderr
from typing import Optional, Tuple

# Handle pytest `ModuleNotFoundError` on non-Windows OS
if system() == "Windows":
    from win32event import CreateEvent, WaitForSingleObject, WAIT_OBJECT_0

from can.interfaces.pcan.basic import (
    PCAN_ERROR_OK,
    PCAN_ERROR_QRCVEMPTY,
    PCAN_RECEIVE_EVENT,
)

from mytoolit.can import SensorConfig
from mytoolit.cmdline import channel_number, mac_address, sth_name
from mytoolit.config import settings
from mytoolit.measurement.acceleration import convert_raw_to_g
from mytoolit.measurement.storage import Storage
from mytoolit.old.network import Network
from mytoolit.old.MyToolItNetworkNumbers import MyToolItNetworkNr
from mytoolit.old.MyToolItCommands import (
    AdcAcquisitionTime,
    AdcOverSamplingRate,
    AdcReference,
    AtvcFormat,
    byte_list_to_int,
    calcSamplingRate,
    DataSets,
    int_to_mac_address,
    MyToolItBlock,
    MyToolItStreaming,
    Prescaler,
    SystemCommandBlueTooth,
    SystemCommandRouting,
)
from mytoolit.old.Plotter import vPlotter, tArray2Binary

Watch = {
    # Time period displayed in the graph in seconds,
    # i.e. “length” of abscissa (“x-coordinate”)
    "DisplayTimeMax": 10,
    # Time after which graph data will be updated again in milliseconds
    "DisplaySampleRateMs": 1 / 5 * 1000,
    # Number of data samples printed in a block,
    # i.e in the time span specified above
    "DisplayBlockSize": 20,
    # Time out after receiving no data in acquiring mode in seconds
    "AliveTimeOutMs": 4000,
}


class CommandLineInterface:
    """ICOc command line interface

    Use this class to connect to the ICOtronic system and acquire measurement
    data using a command line interface.

    """

    def __init__(self, *args, **kwargs):
        # Check if output directory exists and try to create it, if it does not
        # exist already
        try:
            settings.check_output_directory()
        except (NotADirectoryError, OSError) as error:
            raise error

        self.parse_arguments()

        self.logger = getLogger(__name__)
        self.logger.setLevel(self.args.log.upper())
        handler = FileHandler("cli.log", delay=True)
        handler.setFormatter(
            Formatter("{asctime} {levelname} {name} {message}", style="{")
        )
        self.logger.addHandler(handler)
        self.logger.info("Initialized logger")

        self.KeyBoardInterrupt = False
        self.bError = False
        self.iMsgLoss = 0
        self.iMsgsTotal = 0
        self.iMsgCounterLast = -1
        self.Can = Network(
            sender=MyToolItNetworkNr["SPU1"],
            receiver=MyToolItNetworkNr["STH1"],
            log_destination="network.log",
            log_level=self.args.log.upper(),
        )
        self.logger.info("Initialized CAN class")

        channels = (
            self.args.first_channel,
            self.args.second_channel,
            self.args.third_channel,
        )
        self.set_sensors(*channels)
        for channel, value in enumerate(channels, start=1):
            self.logger.info(f"Measurement Channel {channel}: {value}")

        self.connect = (
            True
            if "name" in self.args or "bluetooth_address" in self.args
            else False
        )
        self.sth_name = self.args.name if "name" in self.args else ""
        self.vDeviceAddressSet(
            str(int.from_bytes(self.args.bluetooth_address, "big"))
            if "bluetooth_address" in self.args
            else "0"
        )

        self.vAdcConfig(
            self.args.prescaler, self.args.acquisition, self.args.oversampling
        )
        self.vAdcRefVConfig(self.args.voltage_reference)
        self.vRunTime(0 if self.args.run_time <= 0 else self.args.run_time)
        self.vGraphInit(
            Watch["DisplaySampleRateMs"], Watch["DisplayBlockSize"]
        )
        self.Can.readThreadStop()
        # Set plotter host and port
        self.sPloterSocketHost = settings.gui.host
        self.iPloterSocketPort = settings.gui.port

        self.storage = None
        self.set_output_filename(self.args.filename)

    def __exit__(self):
        if self.storage is not None:
            self.storage.close()
        self.guiProcessStop()
        self.Can.ReadThreadReset()
        if self.Can.bConnected:
            self._BlueToothStatistics()
            ReceiveFailCounter = self._RoutingInformation()
            self._statusWords()
            if ReceiveFailCounter > 0:
                self.bError = True
            self.Can.bBlueToothDisconnect(MyToolItNetworkNr["STU1"])
        if self.Can.bConnected:
            self.Can.readThreadStop()
        self.logger.info("Cleanup done")

        if self.bError:
            self.logger.error("An error occurred")

        self.Can.__exit__()
        self.logger.info("Closed CAN connection")
        if self.bError:
            raise RuntimeError("An error occurred")

    def parse_arguments(self):
        self.parser = argparse.ArgumentParser(
            description="Configure and measure data with the ICOtronic system",
            argument_default=argparse.SUPPRESS,
            formatter_class=ArgumentDefaultsHelpFormatter,
        )

        connection_group = self.parser.add_argument_group(title="Connection")
        connection_group = connection_group.add_mutually_exclusive_group()
        connection_group.add_argument(
            "-b",
            "--bluetooth-address",
            type=mac_address,
            required=False,
            help=(
                "connect to device with specified Bluetooth address "
                "(e.g. “08:6b:d7:01:de:81”)"
            ),
        )
        connection_group.add_argument(
            "-n",
            "--name",
            type=sth_name,
            nargs="?",
            const="",
            required=False,
            help="connect to device with specified name",
        )

        measurement_group = self.parser.add_argument_group(title="Measurement")
        measurement_group.add_argument(
            "-f",
            "--filename",
            type=str,
            default="Measurement",
            required=False,
            help="base name of the output file",
        )

        measurement_group.add_argument(
            "-r",
            "--run-time",
            metavar="SECONDS",
            type=int,
            default=0,
            required=False,
            help=(
                "run time in seconds "
                "(values equal or below “0” specify infinite runtime)"
            ),
        )

        measurement_group.add_argument(
            "-1",
            "--first-channel",
            type=channel_number,
            default=1,
            const=1,
            nargs="?",
            help=(
                "sensor channel number for first measurement channel "
                "(1 - 255; 0 to disable)"
            ),
        )
        measurement_group.add_argument(
            "-2",
            "--second-channel",
            type=channel_number,
            default=0,
            const=2,
            nargs="?",
            help=(
                "sensor channel number for second measurement channel "
                "(1 - 255; 0 to disable)"
            ),
        )
        measurement_group.add_argument(
            "-3",
            "--third-channel",
            type=channel_number,
            default=0,
            const=3,
            nargs="?",
            help=(
                "sensor channel number for third measurement channel "
                "(1 - 255; 0 to disable)"
            ),
        )

        adc_group = self.parser.add_argument_group(title="ADC")

        adc_group.add_argument(
            "-s",
            "--prescaler",
            type=int,
            choices=range(2, 128),
            metavar="2–127",
            default=2,
            required=False,
            help="Prescaler value",
        )
        adc_group.add_argument(
            "-a",
            "--acquisition",
            type=int,
            choices=AdcAcquisitionTime.keys(),
            default=8,
            required=False,
            help="Acquisition time value",
        )
        adc_group.add_argument(
            "-o",
            "--oversampling",
            type=int,
            choices=AdcOverSamplingRate.keys(),
            default=64,
            required=False,
            help="Oversampling rate value",
        )
        adc_group.add_argument(
            "-v",
            "--voltage-reference",
            choices=AdcReference.keys(),
            default="VDD",
            required=False,
            help="Reference voltage",
        )

        logging_group = self.parser.add_argument_group(title="Logging")
        logging_group.add_argument(
            "--log",
            choices=("debug", "info", "warning", "error", "critical"),
            default="info",
            required=False,
            help="Minimum level of messages written to log",
        )

        self.args = self.parser.parse_args()

    def set_output_filename(self, name: Optional[str] = None) -> None:
        """Set the (base) name of the HDF output file

        Parameters
        ----------

        name:
            A new (base) name for the output file

            If you set this parameter to `None`, then the most recently set
            filename will be used

        """

        filename = (
            Path(settings.measurement.output.filename)
            if name is None
            else Path(name)
        )
        if not filename.suffix:
            filename = filename.with_suffix(".hdf5")

        self.output_filename = filename

    def get_output_filepath(self) -> Path:
        """Get the filepath of the HDF output file

        The filepath returned by this method will always include a current
        timestamp to make sure that there are no conflicts with old output
        files.

        Returns
        -------

        The path to the current HDF file

        """

        directory = settings.output_directory()
        filename = self.output_filename
        timestamp = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        filepath = directory.joinpath(
            f"{filename.stem}_{timestamp}{filename.suffix}"
        )

        return filepath

    def _statusWords(self):
        self.logger.info(
            "STH Status Word: {}".format(
                self.Can.node_status(MyToolItNetworkNr["STH1"])
            )
        )
        self.logger.info(
            "STU Status Word: {}".format(
                self.Can.node_status(MyToolItNetworkNr["STU1"])
            )
        )

        status = self.Can.error_status(MyToolItNetworkNr["STH1"])
        if status.adc_overrun():
            self.bError = True
        self.logger.info(f"STH Error Word: {status}")

        self.logger.info(
            "STU Error Word: {}".format(
                self.Can.error_status(MyToolItNetworkNr["STU1"])
            )
        )

    def _BlueToothStatistics(self):
        def log_statistics(node, node_description):
            send_counter = self.Can.BlueToothCmd(
                MyToolItNetworkNr[node], SystemCommandBlueTooth["SendCounter"]
            )
            self.logger.info(
                f"Bluetooth send counter of {node_description}: {send_counter}"
            )
            receive_counter = self.Can.BlueToothCmd(
                MyToolItNetworkNr[node],
                SystemCommandBlueTooth["ReceiveCounter"],
            )
            self.logger.info(
                "Bluetooth receive counter of "
                f"{node_description}: {receive_counter}"
            )

            rssi = self.Can.BlueToothRssi(MyToolItNetworkNr[node])
            self.logger.info(f"RSSI of {node_description}: {rssi} dBm")

        log_statistics("STH1", "sensor node")
        log_statistics("STU1", "STU")

    def _RoutingInformationSthSend(self):
        self.iSthSendCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STH1"],
            SystemCommandRouting["SendCounter"],
            MyToolItNetworkNr["STU1"],
        )
        self.logger.info(
            f"Sensor node - Send counter (port STU): {self.iSthSendCounter}"
        )
        SendCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STH1"],
            SystemCommandRouting["SendFailCounter"],
            MyToolItNetworkNr["STU1"],
        )
        self.logger.info(
            f"Sensor node - Send fail counter (port STU): {SendCounter}"
        )
        SendCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STH1"],
            SystemCommandRouting["SendLowLevelByteCounter"],
            MyToolItNetworkNr["STU1"],
        )
        self.logger.info(
            f"Sensor node - Send byte counter (port STU): {SendCounter}"
        )

    def _RoutingInformationSthReceive(self):
        ReceiveCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STH1"],
            SystemCommandRouting["ReceiveCounter"],
            MyToolItNetworkNr["STU1"],
        )
        self.logger.info(
            f"Sensor node - Receive counter (port STU): {ReceiveCounter}"
        )
        ReceiveFailCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STH1"],
            SystemCommandRouting["ReceiveFailCounter"],
            MyToolItNetworkNr["STU1"],
        )
        self.logger.info(
            "Sensor node - Receive fail counter (port STU): "
            f"{ReceiveFailCounter}"
        )
        ReceiveCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STH1"],
            SystemCommandRouting["ReceiveLowLevelByteCounter"],
            MyToolItNetworkNr["STU1"],
        )
        self.logger.info(
            f"Sensor node - Receive byte counter (port STU): {ReceiveCounter}"
        )
        return ReceiveFailCounter

    def _RoutingInformationSth(self):
        self._RoutingInformationSthSend()
        ReceiveFailCounter = self._RoutingInformationSthReceive()
        return ReceiveFailCounter

    def _RoutingInformationStuPortSpuSend(self):
        self.iStuSendCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["SendCounter"],
            MyToolItNetworkNr["SPU1"],
        )
        self.logger.info(
            "STU1 - Send Counter(Port SPU1): " + str(self.iStuSendCounter)
        )
        SendCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["SendFailCounter"],
            MyToolItNetworkNr["SPU1"],
        )
        self.logger.info(
            "STU1 - Send Fail Counter(Port SPU1): " + str(SendCounter)
        )
        SendCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["SendLowLevelByteCounter"],
            MyToolItNetworkNr["SPU1"],
        )
        self.logger.info(
            "STU1 - Send Byte Counter(Port SPU1): " + str(SendCounter)
        )

    def _RoutingInformationStuPortSpuReceive(self):
        ReceiveCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["ReceiveCounter"],
            MyToolItNetworkNr["SPU1"],
        )
        self.logger.info(
            "STU1 - Receive Counter(Port SPU1): " + str(ReceiveCounter)
        )
        ReceiveFailCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["ReceiveFailCounter"],
            MyToolItNetworkNr["SPU1"],
        )
        self.logger.info(
            "STU1 - Receive Fail Counter(Port SPU1): "
            + str(ReceiveFailCounter)
        )
        ReceiveCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["ReceiveLowLevelByteCounter"],
            MyToolItNetworkNr["SPU1"],
        )
        self.logger.info(
            "STU1 - Receive Byte Counter(Port SPU1): " + str(ReceiveCounter)
        )
        return ReceiveFailCounter

    def _RoutingInformationStuPortSpu(self):
        self._RoutingInformationStuPortSpuSend()
        ReceiveFailCounter = self._RoutingInformationStuPortSpuReceive()
        return ReceiveFailCounter

    def _RoutingInformationStuPortSthSend(self):
        iStuSendCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["SendCounter"],
            MyToolItNetworkNr["STH1"],
        )
        self.logger.info(
            "STU1 - Send Counter(Port STH1): " + str(iStuSendCounter)
        )
        SendCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["SendFailCounter"],
            MyToolItNetworkNr["STH1"],
        )
        self.logger.info(
            "STU1 - Send Fail Counter(Port STH1): " + str(SendCounter)
        )
        SendCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["SendLowLevelByteCounter"],
            MyToolItNetworkNr["STH1"],
        )
        self.logger.info(
            "STU1 - Send Byte Counter(Port STH1): " + str(SendCounter)
        )

    def _RoutingInformationStuPortSthReceive(self):
        ReceiveCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["ReceiveCounter"],
            MyToolItNetworkNr["STH1"],
        )
        self.logger.info(
            "STU1 - Receive Counter(Port STH1): " + str(ReceiveCounter)
        )
        ReceiveFailCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["ReceiveFailCounter"],
            MyToolItNetworkNr["STH1"],
        )
        self.logger.info(
            "STU1 - Receive Fail Counter(Port STH1): "
            + str(ReceiveFailCounter)
        )
        ReceiveCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["ReceiveLowLevelByteCounter"],
            MyToolItNetworkNr["STH1"],
        )
        self.logger.info(
            "STU1 - Receive Byte Counter(Port STH1): " + str(ReceiveCounter)
        )
        return ReceiveFailCounter

    def _RoutingInformationStuPortSth(self):
        self._RoutingInformationStuPortSthSend()
        ReceiveFailCounter = self._RoutingInformationStuPortSthReceive()
        return ReceiveFailCounter

    def _RoutingInformation(self):
        ReceiveFailCounter = self._RoutingInformationSth()
        ReceiveFailCounter += self._RoutingInformationStuPortSth()
        ReceiveFailCounter += self._RoutingInformationStuPortSpu()
        iSendFail = self.iSthSendCounter - self.iStuSendCounter
        if 0 > iSendFail:
            iSendFail = 0
        iSendFail = iSendFail / self.iSthSendCounter
        iSendFail *= 100
        self.logger.info("Send fail approximately: " + str(iSendFail) + "%")
        if 0 < iSendFail:
            print("Send fail approximately: " + str(iSendFail) + "%")
        return ReceiveFailCounter

    def set_sensors(self, first: int, second: int, third: int) -> None:
        """Set sensor numbers for measurement channels

        Parameters
        ----------

        first:
            Sensor number for first measurement channel

        second:
            Sensor number for second measurement channel

        third:
            Sensor number for third measurement channel

        """

        self.sensor = SensorConfig(first, second, third)

        dataSets = self.Can.data_sets(*map(bool, (first, second, third)))
        self.tAccDataFormat = DataSets[dataSets]

    def vDeviceAddressSet(self, iAddress):
        """Set bluetooth device address"""
        iAddress = int(iAddress, base=0)
        if 0 < iAddress and iAddress < 2**48 - 1:
            iAddress = hex(iAddress)
            self.iAddress = iAddress
        else:
            self.iAddress = 0

    def vAdcConfig(self, iPrescaler, iAquistionTime, iOversampling):
        if Prescaler["Min"] > iPrescaler:
            iPrescaler = Prescaler["Min"]
        elif Prescaler["Max"] < iPrescaler:
            iPrescaler = Prescaler["Max"]

        iAcquisitionTime = AdcAcquisitionTime[iAquistionTime]
        iOversampling = AdcOverSamplingRate[iOversampling]
        self.samplingRate = int(
            calcSamplingRate(iPrescaler, iAcquisitionTime, iOversampling) + 0.5
        )
        self.iPrescaler = iPrescaler
        self.iAquistionTime = iAcquisitionTime
        self.iOversampling = iOversampling

    def vAdcRefVConfig(self, sAdcRef):
        self.sAdcRef = sAdcRef

    def vRunTime(self, runTime):
        self.iRunTime = int(runTime)

    def vGraphInit(self, sampleInterval=200, blockSize=10):
        """
        sampleInterval in ms
        """

        self.tDataPointTimeStamp = 0  # Time stamp of last data point
        self.iPacketLossTimeStamp = 0
        self.iGraphBlockSize = blockSize
        self.iGraphSampleInterval = sampleInterval
        self.sMsgLoss = "Acceleration(" + str(format(0, "3.3f")) + "%)"
        self.GuiPackage = {"X": [], "Y": [], "Z": []}

    def vStuAddr(self, sStuAddr):
        self.sStuAddr = sStuAddr

    def guiProcessStop(self):
        if hasattr(self, "tSocket"):
            try:
                self.vGraphSend(["Run", False])
            except (ConnectionAbortedError, OSError):
                pass
            self.tSocket.close()
        if hasattr(self, "guiProcess"):
            self.guiProcess.terminate()
            self.guiProcess.join()

    def vGraphSend(self, data):
        bSend = True
        data = tArray2Binary(data)
        while bSend:
            self.tSocket.sendall(data)
            sleep(0.1)
            ack = self.tSocket.recv(2**10)
            self.logger.debug(f"Received acknowledgment: {ack}")
            if ack is not None and ack == data:
                bSend = False

    def guiProcessRestart(self):
        self.guiProcessStop()

        self.guiProcess = multiprocessing.Process(
            target=vPlotter,
            args=(self.iPloterSocketPort, self.logger.getEffectiveLevel()),
        )
        self.guiProcess.start()

        # Wait until socket of GUI application is ready
        connection_established = False
        while not connection_established:
            try:
                self.logger.debug("Try to initialize socket")
                self.tSocket = socket.socket(
                    socket.AF_INET, socket.SOCK_STREAM
                )
                self.tSocket.connect(
                    (self.sPloterSocketHost, self.iPloterSocketPort)
                )
                connection_established = True
            except ConnectionError:
                sleep(0.1)

        self.logger.debug("Initialized socket")

        self.vGraphSend(["dataBlockSize", self.iGraphBlockSize])
        self.vGraphSend(["sampleInterval", self.iGraphSampleInterval])
        self.vGraphSend(["xDim", Watch["DisplayTimeMax"]])
        self.update_packet_loss()
        if self.sensor.first:
            self.vGraphSend(["lineNameX", "Signal 1"])
        if self.sensor.second:
            self.vGraphSend(["lineNameY", "Signal 2"])
        if self.sensor.third:
            self.vGraphSend(["lineNameZ", "Signal 3"])
        self.vGraphSend(["Plot", True])

    def update_graph_data(self, x=0, y=0, z=0):
        timeStampNow = int(round(time() * 1000))
        elapsed_time_ms = timeStampNow - self.tDataPointTimeStamp
        # Only add a single data sample for each part of the current block
        if elapsed_time_ms <= self.iGraphSampleInterval / self.iGraphBlockSize:
            return

        self.tDataPointTimeStamp = timeStampNow
        self.GuiPackage["X"].append(x)
        self.GuiPackage["Y"].append(y)
        self.GuiPackage["Z"].append(z)
        # Send the data to the plotter, after we collected all samples for the
        # current block
        if len(self.GuiPackage["X"]) >= self.iGraphBlockSize:
            try:
                self.tSocket.sendall(tArray2Binary(["data", self.GuiPackage]))
            except (ConnectionAbortedError, ConnectionResetError):
                # Closing the plotter window quits the plotter process and
                # there might be not socket to send data to after that
                pass

            self.GuiPackage = {"X": [], "Y": [], "Z": []}

    def update_packet_loss(self, msgCounter=-1):
        if self.iMsgCounterLast == -1:
            self.iMsgCounterLast = msgCounter
        else:
            self.iMsgCounterLast += 1
            self.iMsgCounterLast %= 256
        if self.iMsgCounterLast != msgCounter:
            iLost = (
                msgCounter - self.iMsgCounterLast
                if msgCounter > self.iMsgCounterLast
                else 0xFF - self.iMsgCounterLast + msgCounter
            )
            self.iMsgLoss += iLost
            self.iMsgsTotal += iLost
            if 0 > iLost:
                self.iMsgLoss += 256
                self.iMsgsTotal += 256
            self.iMsgCounterLast = msgCounter
        else:
            self.iMsgsTotal += 1
        iPacketLossTimeStamp = int(round(time() * 1000))
        time_since_update = iPacketLossTimeStamp - self.iPacketLossTimeStamp
        if time_since_update >= 1000:
            self.iPacketLossTimeStamp = iPacketLossTimeStamp
            pakets_received = (1 - self.iMsgLoss / self.iMsgsTotal) * 100
            sMsgLoss = f"Signal Quality: {pakets_received:3.2f} %"
            if sMsgLoss != self.sMsgLoss:
                self.sMsgLoss = sMsgLoss
                try:
                    self.tSocket.sendall(
                        tArray2Binary(["diagramName", self.sMsgLoss])
                    )
                except (ConnectionAbortedError, ConnectionResetError):
                    # Closing the plotter window quits the plotter process and
                    # there might be no socket to send data to after that
                    pass

            self.iMsgLoss = 0
            self.iMsgsTotal = 0

    def reset(self):
        if self.KeyBoardInterrupt:
            return

        try:
            self.Can.ReadThreadReset()
            self.Can.reset_node("STU1")
            self.vStuAddr(
                int_to_mac_address(
                    self.Can.BlueToothAddress(MyToolItNetworkNr["STU1"])
                )
            )
            self.guiProcessStop()
        except KeyboardInterrupt:
            self.KeyBoardInterrupt = True

    def read_acceleration_range(self) -> Tuple[int, bool]:
        """Read the range of the acceleration sensor

        Returns
        -------

        A tuple containing

        - the acceleration range in multiples of g₀ or a default value for a ±
          100 g₀ sensor, if reading the EEPROM value was unsuccessful,
        - a boolean that specifies if reading was successful (`True`) or not
          (`False`).

        """

        # Default values
        acceleration_range_g = 200
        success = False

        try:
            acceleration_range_g = int(
                self.Can.read_acceleration_sensor_range_in_g()
            )
            success = True
        except ValueError:
            pass

        return (acceleration_range_g, success)

    def update_sensor_config(self) -> None:
        """Update sensor configuration in the connected sensor device"""

        # We use special sensor channel number `0` (currently defined as “no
        # change to channel number”) as default value.
        #
        # This should not be a problem in connection with the definition in the
        # CLI interface of ICOc, where `0` is defined as disabling the
        # measurement channel. Here we would need to enter the sensor channel
        # number to enable the measurement anyway.
        self.Can.write_sensor_config(
            *[
                0 if sensor is None or sensor <= 0 else sensor
                for sensor in (
                    self.sensor.first,
                    self.sensor.second,
                    self.sensor.third,
                )
            ]
        )

    def vDataAquisition(self):
        if self.KeyBoardInterrupt:
            return

        try:
            if self.Can.bConnected:
                self.Can.ConfigAdc(
                    MyToolItNetworkNr["STH1"],
                    self.iPrescaler,
                    self.iAquistionTime,
                    self.iOversampling,
                    AdcReference[self.sAdcRef],
                )
                # Initialize HDF output
                self.storage = Storage(self.get_output_filepath())
                # We need the acceleration range later to convert the ADC
                # acceleration values into multiples of g₀
                (
                    self.acceleration_range_g,
                    success,
                ) = self.read_acceleration_range()
                if not success:
                    print(
                        (
                            "Warning: Unable to determine sensor range from "
                            "EEPROM value — Assuming ± 100 g sensor"
                        ),
                        file=stderr,
                    )
                if self.acceleration_range_g < 1:
                    print(
                        (
                            "Warning: Sensor range"
                            f" “{self.acceleration_range_g}” below 1 g — Using"
                            " range 200 instead (± 100 g sensor)"
                        ),
                        file=stderr,
                    )
                    self.acceleration_range_g = 200

                # ICOc does not use the network class to read the streaming
                # data but uses the `Read` method of the PCAN API directly.
                # This means we have to disable the read thread of the CAN
                # class first.
                self.Can.readThreadStop()
                self.guiProcessRestart()
                self.logger.info("Start Acquiring Data")
                self.vGetStreamingAccData()
            else:
                self.logger.error("Device not allocable")
        except KeyboardInterrupt:
            self.KeyBoardInterrupt = True
            self.__exit__()

    def close(self):
        if self.Can.RunReadThread:
            self.__exit__()

    def get_can_error_message(
        self, status: int, prefix: Optional[str] = None
    ) -> str:
        """Retrieve a human readable CAN error message

        Parameters
        ----------

        status:
            The status number returned by a call to the PCAN-Basic API

        prefix:
            An optional description of the error condition

        Returns
        -------

        A textual description of the CAN error

        """

        error_message = self.Can.pcan.GetErrorText(status)[1].decode()
        description = "" if prefix is None else "{prefix}: "
        return f"{description}{error_message}"

    def read_streaming(self):
        """Read streaming messages"""
        self.logger.debug("Add CAN read event")
        receive_event = CreateEvent(None, 0, 0, None)
        status = self.Can.pcan.SetValue(
            self.Can.m_PcanHandle, PCAN_RECEIVE_EVENT, int(receive_event)
        )

        if status != PCAN_ERROR_OK:
            error_message = self.get_can_error_message(
                status, "Unable to set CAN receive event"
            )
            raise Exception(error_message)

        TIMEOUT_SECONDS = 4
        time_last_read = time()
        time_since_read = 0
        self.logger.debug("Wait for CAN data")
        while (
            self.Can.get_elapsed_time() < self.aquireEndTime
            and time_since_read <= TIMEOUT_SECONDS
            and self.guiProcess.is_alive()
        ):
            if WaitForSingleObject(receive_event, 50) == WAIT_OBJECT_0:
                self.read_streaming_messages()
                time_last_read = time()
            time_since_read = time() - time_last_read

        if time_since_read >= TIMEOUT_SECONDS:
            print(
                (
                    "Exiting program since no streaming data was received "
                    f"in the last {round(time_since_read, 2)} seconds"
                ),
                file=stderr,
            )

        self.Can.pcan.SetValue(self.Can.m_PcanHandle, PCAN_RECEIVE_EVENT, 0)

    def read_streaming_messages(self):
        """Read multiple streaming messages"""
        status = PCAN_ERROR_OK

        self.logger.debug("Read streaming messages")

        while status != PCAN_ERROR_QRCVEMPTY:
            status = self.read_streaming_message()
            if status not in {PCAN_ERROR_OK, PCAN_ERROR_QRCVEMPTY}:
                error_message = self.get_can_error_message(
                    status, "Unable to read streaming message"
                )
                raise Exception(error_message)

    def read_streaming_message(self):
        """Read single streaming message"""

        status, message, timestamp = self.Can.pcan.Read(self.Can.m_PcanHandle)
        if status == PCAN_ERROR_OK:
            timestamp_ms = (
                timestamp.millis_overflow * (2**32)
                + timestamp.millis
                + timestamp.micros / 1000
            )
            if message.ID == self.AccAckExpected.ID:
                self.update_packet_loss(message.DATA[1])
                self.update_acceleration_data(message.DATA, timestamp_ms)

        return status

    def vGetStreamingAccDataProcess(self):
        try:
            # Read streaming messages until
            # - plotter window is closed or
            # - runtime is up.
            self.read_streaming()

            # We store the acceleration metadata at the end since, at this time
            # the acceleration table should exist. Otherwise we would not be
            # able to add the metadata to the table.
            sensor_range = self.acceleration_range_g / 2
            self.storage.add_acceleration_meta(
                "Sensor_Range", f"± {sensor_range} g₀"
            )

            self.__exit__()
        except KeyboardInterrupt:
            self.KeyBoardInterrupt = True
            print("Data acquisition terminated")
            self.__exit__()

    def vGetStreamingAccDataAccStart(self):
        if not any((self.sensor.first, self.sensor.second, self.sensor.third)):
            return True

        ack = None
        accFormat = AtvcFormat()
        accFormat.asbyte = 0
        accFormat.b.bStreaming = 1
        accFormat.b.bNumber1 = int(bool(self.sensor.first))
        accFormat.b.bNumber2 = int(bool(self.sensor.second))
        accFormat.b.bNumber3 = int(bool(self.sensor.third))
        accFormat.b.u3DataSets = self.tAccDataFormat
        cmd = self.Can.CanCmd(
            MyToolItBlock["Streaming"], MyToolItStreaming["Data"], 0, 0
        )
        self.AccAckExpected = self.Can.CanMessage20(
            cmd,
            MyToolItNetworkNr["STH1"],
            MyToolItNetworkNr["SPU1"],
            [accFormat.asbyte],
        )
        cmd = self.Can.CanCmd(
            MyToolItBlock["Streaming"], MyToolItStreaming["Data"], 1, 0
        )
        message = self.Can.CanMessage20(
            cmd,
            MyToolItNetworkNr["SPU1"],
            MyToolItNetworkNr["STH1"],
            [accFormat.asbyte],
        )
        self.logger.info(
            "MsgId/Subpayload(Acc): "
            + hex(message.ID)
            + "/"
            + hex(accFormat.asbyte)
        )
        endTime = self.Can.get_elapsed_time() + 4000
        while ack is None and self.Can.get_elapsed_time() < endTime:
            self.Can.WriteFrame(message)
            readEndTime = self.Can.get_elapsed_time() + 500
            while ack is None and self.Can.get_elapsed_time() < readEndTime:
                ack = self.ReadMessage()
        return ack

    def vGetStreamingAccData(self):
        ack = self.vGetStreamingAccDataAccStart()
        currentTime = self.Can.get_elapsed_time()
        if ack is None:
            self.logger.error(
                f"No acknowledge received from STH “{self.iAddress}”"
            )
            self.aquireEndTime = currentTime
        elif self.iRunTime == 0:
            self.aquireEndTime = currentTime + (1 << 32)
        else:
            self.aquireEndTime = currentTime + self.iRunTime * 1000
        self.vGetStreamingAccDataProcess()

    def update_acceleration_data(self, data, timestamp_ms):
        timestamp = round(timestamp_ms, 3)
        counter = data[1]

        axes = [
            axis
            for axis, activated in (
                ("x", self.sensor.first),
                ("y", self.sensor.second),
                ("z", self.sensor.third),
            )
            if activated
        ]

        if len(axes) <= 0:
            return

        convert_acceleration = partial(
            convert_raw_to_g, max_value=self.acceleration_range_g
        )
        number_values = 3 if self.tAccDataFormat == DataSets[3] else len(axes)
        values = [
            convert_acceleration(
                byte_list_to_int(data[start : start + 2])
            ).magnitude
            for start in range(2, 2 + number_values * 2, 2)
        ]

        if self.tAccDataFormat == DataSets[1]:
            axis_values = {axis: value for axis, value in zip(axes, values)}
            self.storage.add_acceleration(
                values=axis_values, counter=counter, timestamp=timestamp
            )
            self.update_graph_data(**axis_values)
        elif self.tAccDataFormat == DataSets[3]:
            axis = axes[0]
            for value in values:
                self.storage.add_acceleration(
                    values={axis: value}, counter=counter, timestamp=timestamp
                )
            self.update_graph_data(**{axis: values[0]})
        else:
            self.logger.error("Wrong Ack format")

    def ReadMessage(self):
        status, message, timestamp = self.Can.pcan.Read(self.Can.m_PcanHandle)
        if status == PCAN_ERROR_OK:
            peakCanTimeStamp = (
                timestamp.millis_overflow * (2**32)
                + timestamp.millis
                + timestamp.micros / 1000
            )
            result = {
                "CanMsg": message,
                "PcTime": self.Can.get_elapsed_time(),
                "PeakCanTime": peakCanTimeStamp,
            }
            return result

        if status != PCAN_ERROR_QRCVEMPTY:
            error_message = self.get_can_error_message(
                status, "Unexpected CAN status value"
            )
            self.logger.error(error_message)
            print(error_message, file=stderr)
            raise Exception(error_message)

        return None

    def _vRunConsoleStartupLoggerPrint(self):
        self.logger.info(f"STH Name: {self.sth_name}")
        self.logger.info(f"Bluetooth Address: {self.iAddress}")
        self.logger.info(f"Connect to STH: {self.connect}")
        self.logger.info(f"Run Time: {self.iRunTime} s")
        self.logger.info(f"Prescaler: {self.iPrescaler}")
        aqcuisition_time = AdcAcquisitionTime.inverse[self.iAquistionTime]
        oversampling_rate = AdcOverSamplingRate.inverse[self.iOversampling]
        self.logger.info(f"Acquisition Time: {aqcuisition_time}")
        self.logger.info(f"Oversampling Rate: {oversampling_rate}")
        self.logger.info(f"Reference Voltage: {self.sAdcRef}")
        data_sets = DataSets.inverse[self.tAccDataFormat]
        self.logger.info(f"Data Sets: {data_sets}")
        self.logger.info(f"Sensors: {self.sensor}")

    def _vRunConsoleStartup(self):
        self._vRunConsoleStartupLoggerPrint()

    def vRunConsoleAutoConnect(self):
        if self.iAddress not in {0, "0", "0x0"}:
            self.Can.bBlueToothConnectPollingAddress(
                MyToolItNetworkNr["STU1"], self.iAddress
            )
        else:
            self.Can.bBlueToothConnectPollingName(
                MyToolItNetworkNr["STU1"], self.sth_name, log=False
            )
        if self.Can.bConnected:
            self.update_sensor_config()
            self.vDataAquisition()

    def run(self):
        self._vRunConsoleStartup()
        self.reset()
        if self.connect:
            self.vRunConsoleAutoConnect()
        self.close()


if __name__ == "__main__":
    CommandLineInterface().run()



# -- Imports ------------------------------------------------------------------

from unittest import TextTestRunner

from .extended_test_result import ExtendedTestResult

# -- Class --------------------------------------------------------------------


class ExtendedTestRunner(TextTestRunner):
    """Extend default test runner to change result class"""

    def __init__(self, *arguments, **keyword_arguments):
        """Initialize the test runner"""

        keyword_arguments["resultclass"] = ExtendedTestResult
        super().__init__(*arguments, **keyword_arguments)



def testRampDim(MaxValue, FirstDim, SecondDim, fRecalc):
    ramp = []

    u64Quantisation = MaxValue / (FirstDim * SecondDim - 1)
    if None != fRecalc:
        for i in range(0, FirstDim * SecondDim):
            ramp.append(fRecalc(int(u64Quantisation * i)))
    else:
        for i in range(0, FirstDim * SecondDim):
            ramp.append(int(u64Quantisation * i))
    return ramp



# -- Imports ------------------------------------------------------------------

from typing import List, Optional, Union

from mytoolit.old.MyToolItCommands import NetworkState, NodeState

# -- Classes ------------------------------------------------------------------


class State:
    """Wrapper class for the byte returned by the Get/Set State command

    See also: https://mytoolit.github.io/Documentation/#command:get-set-state

    """

    def __init__(
        self,
        *value: int,
        mode: Optional[str] = None,
        location: Union[None, int, str] = None,
        state: Union[None, int, str] = None,
    ) -> None:
        """Initialize the node status word using the given arguments

        Parameters
        -----------

        value:
            The value of the state byte

        mode:
            Specifies if the state should be set or retrieved (get)

        location:
            A string or number that specifies the code location

        state:
            A string or number that specifies the network state

        Examples
        --------

        >>> State(mode='Get', location='Bootloader', state='Operating')
        Get State, Location: Bootloader, State: Operating

        >>> State(mode='Set', state='Operating').value == 0b1_0000_101
        True

        """

        def set_part(start, width, number):
            """Store bit pattern number at bit start of the identifier"""

            state_ones = 0xFF
            mask = (1 << width) - 1

            # Set all bits for targeted part to 0
            self.value &= (mask << start) ^ state_ones
            # Make sure we use the correct number of bits for number
            number = number & mask
            # Set bits to given value
            self.value |= number << start

        self.value = value[0] if value else 0

        # ========
        # = Mode =
        # ========

        if mode is not None:
            if mode not in {"Get", "Set"}:
                raise ValueError(f"Unknown mode “{mode}”")

            set_part(start=7, width=1, number=int(mode == "Set"))

        # ============
        # = Location =
        # ============

        if isinstance(location, str):
            try:
                location = NodeState[location]
            except KeyError:
                raise ValueError(f"Unknown location “{location}”")

        if location is not None:
            set_part(start=4, width=2, number=location)

        # =========
        # = State =
        # =========

        if isinstance(state, str):
            try:
                state = NetworkState[state]
            except KeyError:
                raise ValueError(f"Unknown state “{state}”")

        if state is not None:
            set_part(start=0, width=3, number=state)

    def __eq__(self, other: object) -> bool:
        """Compare this state to another object

        Parameters
        ----------

        other:
            The other object this state should be compared to

        Returns
        -------

        - True, if the given object is a state and it has the same
          value as this state

        - False, otherwise

        Examples
        --------

        >>> state1 = State(mode='Get',
        ...                location='Bootloader',
        ...                state='Operating')
        >>> state2 = State(mode='Set',
        ...                location='Bootloader',
        ...                state='Operating')

        >>> state1 == state2
        False

        >>> state1 == State(state2.value, mode='Get')
        True

        """

        if isinstance(other, State):
            return self.value == other.value

        return False

    def is_set(self) -> bool:
        """Check if the status should be set

        If this method returns `False`, then the state should be retrieved
        (get) instead.

        Examples
        --------

        >>> State(0b1_0000000).is_set()
        True
        >>> State(0b0_0000000).is_set()
        False
        >>> State(mode='Get').is_set()
        False
        >>> State(mode='Set').is_set()
        True

        """

        return bool((self.value >> 7) & 1)

    def location_name(self) -> str:
        """Retrieve the name of the current (code) location

        Returns
        -------

        The name of the node state represented by this state object

        Examples
        --------

        >>> State(0b00_1010).location_name()
        'No Change'
        >>> State(0b01_1010).location_name()
        'Bootloader'
        >>> State(location='Application').location_name()
        'Application'

        """

        location = (self.value >> 4) & 0b11
        return NodeState.inverse[location]

    def state_name(self) -> str:
        """Retrieve the name of the network state

        Returns
        -------

        The name of the network state represented by this state object

        Examples
        --------

        >>> State(0b000).state_name()
        'Failure'
        >>> State(0b101).state_name()
        'Operating'
        >>> State(state='Startup').state_name()
        'Startup'

        """

        network_state = self.value & 0b111
        return NetworkState.inverse[network_state]

    def __repr__(self) -> str:
        """Retrieve the textual representation of the state

        Returns
        -------

        A string that describes the attributes of the state

        Examples
        --------

        >>> State(0b1_0_01_0_110)
        Set State, Location: Bootloader, State: Startup
        >>> State(0b0_1_11_1_001)
        Get State, Location: Reserved, State: Error

        """

        attributes = [
            "{} State".format("Set" if self.is_set() else "Get"),
            f"Location: {self.location_name()}",
            f"State: {self.state_name()}",
        ]

        return ", ".join(attributes)


class NodeStatus:
    """Wrapper class for the node status word

    Please do not use this class directly, but instead use one of the
    two specific status classes for the STH and STU.

    """

    def __init__(self, value: Union[List[int], int]) -> None:
        """Initialize the node status word using the given arguments

        Parameters
        ----------

        value:
            A 32 bit integer or list of bytes that specifies the value of the
            node status word

        """

        # Currently only the first byte (of the little endian version) of
        # status word 0 contains (non-reserved) data
        self.value = value if isinstance(value, int) else value[0]

    def __repr__(self) -> str:
        """Retrieve the textual representation of the node status word

        Returns
        -------

        A string that describes the attributes of the node status word

        Examples
        --------

        >>> NodeStatus(0b1010)
        State: Operating, No Error

        >>> NodeStatus([0b1010, 0, 0, 0])
        State: Operating, No Error

        >>> NodeStatus(0b1)
        State: Failure, Error

        """

        attributes = [
            f"State: {self.state_name()}",
            f"{'' if self.error() else 'No '}Error",
        ]

        return ", ".join(attributes)

    def error(self) -> bool:
        """Retrieve the status of the error bit

        Returns
        -------

        True if the error bit was set or False otherwise

        Examples
        --------

        >>> NodeStatus(0b0).error()
        False

        >>> NodeStatus(0b1).error()
        True

        """

        return bool(self.value & 1)

    def state_name(self) -> str:
        """Get the name of the state represented by the node status word

        Returns
        -------

        A textual representation of the current node state

        Examples
        --------

        >>> NodeStatus(0b1010).state_name()
        'Operating'

        >>> NodeStatus(0b1110).state_name()
        'No Change'

        """

        state = (self.value >> 1) & 0b111
        return NetworkState.inverse[state]


class NodeStatusSTH(NodeStatus):
    def __init__(self, value: Union[List[int], int]) -> None:
        """Initialize the node status word using the given arguments

        Parameters
        ----------

        value:
            A 32 bit integer or list of bytes that specifies the value of the
            node status word

        """

        super().__init__(value)

    def __repr__(self) -> str:
        """Retrieve the textual representation of the node status word

        Returns
        -------

        A string that describes the attributes of the node status word

        Examples
        --------

        >>> NodeStatusSTH(0b0100)
        State: Standby, No Error

        """

        return super().__repr__()


class NodeStatusSTU(NodeStatus):
    def __init__(self, value: Union[List[int], int]) -> None:
        """Initialize the node status word using the given arguments

        Parameters
        ----------

        value:
            A 32 bit integer or list of bytes that specifies the value of the
            node status word

        """

        super().__init__(value)

    def __repr__(self) -> str:
        """Retrieve the textual representation of the node status word

        Returns
        -------

        A string that describes the attributes of the node status word

        Example
        -------

        >>> NodeStatusSTU(0b1101010) # doctest:+NORMALIZE_WHITESPACE
        State: Operating, No Error, Radio Port Disabled,
        CAN Port Enabled, Bluetooth Connected

        """

        radio_port_enabled = self.value >> 4 & 1
        can_port_enabled = self.value >> 5 & 1
        bluetooth_connected = self.value >> 6 & 1

        attributes = [
            super().__repr__(),
            "Radio Port {}".format(
                "Enabled" if radio_port_enabled else "Disabled"
            ),
            "CAN Port {}".format(
                "Enabled" if can_port_enabled else "Disabled"
            ),
            "Bluetooth {}".format(
                "Connected" if bluetooth_connected else "Disconnected"
            ),
        ]

        return ", ".join(attributes)


class ErrorStatus:
    """Wrapper class for the error status word

    Please do not use this class directly, but instead use one of the
    two specific error status classes for the STH and STU.
    """

    def __init__(self, value: Union[List[int], int]) -> None:
        """Initialize the error status word using the given arguments

        Parameters
        ----------

        value:
            A 32 bit integer or list of bytes that specifies the value of the
            error status word

        """

        # Currently only the first byte (of the little endian version) of
        # status word contains (non-reserved) data
        self.value = value if isinstance(value, int) else value[0]

    def transmission_error(self) -> bool:
        """Retrieve the status of the transmission error bit

        Returns
        -------

        True if the error bit was set or False otherwise

        Examples
        --------

        >>> ErrorStatus(0b0).transmission_error()
        False

        >>> ErrorStatus(0b1).transmission_error()
        True

        """

        return bool(self.value & 1)


class ErrorStatusSTH(ErrorStatus):
    """Wrapper class for error status word 1 of the STH"""

    def __init__(self, value: Union[List[int], int]) -> None:
        """Initialize the error status word using the given arguments

        Parameters
        ----------

        value:
            A 32 bit integer or list of bytes that specifies the value of the
            error status word

        """
        super().__init__(value)

    def __repr__(self) -> str:
        """Retrieve the textual representation of the error status word

        Returns
        -------

        A string that describes the attributes of the error status word

        Examples
        --------

        >>> ErrorStatusSTH(0b0)
        No Error

        >>> ErrorStatusSTH(0b11)
        Bluetooth Transmission Error, ADC Overrun Error

        >>> ErrorStatusSTH(0b10)
        ADC Overrun Error

        """

        errors = []

        if self.transmission_error():
            errors.append("Bluetooth Transmission Error")

        if self.adc_overrun():
            errors.append("ADC Overrun Error")

        if not errors:
            return "No Error"

        return ", ".join(errors)

    def adc_overrun(self) -> bool:
        """Retrieve the status of the ADC overrun bit

        Returns
        -------

        True if the ADC overrun error bit is set or False otherwise

        Examples
        --------

        >>> ErrorStatusSTH(0b10).adc_overrun()
        True

        >>> ErrorStatusSTH(0b11).adc_overrun()
        True

        >>> ErrorStatusSTH(0b01).adc_overrun()
        False

        """

        return bool((self.value >> 1) & 1)


class ErrorStatusSTU(ErrorStatus):
    """Wrapper class for error status word 1 of the STH"""

    def __init__(self, value: Union[List[int], int]) -> None:
        """Initialize the error status word using the given arguments

        Parameters
        ----------

        value:
            A 32 bit integer or list of bytes that specifies the value of the
            error status word

        """
        super().__init__(value)

    def __repr__(self) -> str:
        """Retrieve the textual representation of the error status word

        Returns
        -------

        A string that describes the attributes of the error status word

        Examples
        --------

        >>> ErrorStatusSTU(0b0)
        No Error

        >>> ErrorStatusSTU(0b1)
        CAN Transmission Error

        """

        if self.transmission_error():
            return "CAN Transmission Error"

        return "No Error"


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



# -- Imports ------------------------------------------------------------------

from typing import List, Optional, Union

from bidict import bidict

# -- Class --------------------------------------------------------------------


class CalibrationMeasurementFormat:
    """This class stores message data for the calibration measurement command

    See also:
    https://mytoolit.github.io/Documentation/#command-calibration-measurement

    """

    methods = ["Reserved", "Activate", "Deactivate", "Measure"]
    elements = bidict(
        {
            "Data": 0,
            "Temperature": 1,
            "Voltage": 32,
            "VSS": 96,
            "VDD": 97,
            "Internal": 98,
            "OPV": 99,
        }
    )

    def __init__(
        self,
        *data: Union[bytearray, List[int]],
        set: Optional[bool] = None,
        element: Optional[str] = None,
        method: Optional[str] = None,
        dimension: Optional[int] = None,
        reference_voltage: Optional[float] = None,
    ) -> None:
        """Initialize the calibration measurement format

        This class allows you to specify the message bytes directly as first
        argument. Alternatively you can use the other arguments of the
        initializer to  either

        - overwrite specific parts of the given message bytes or
        - create a calibration format without specifying the message bytes.

        Parameters
        ----------

        data:
            A list containing the (first four) bytes of the calibration
            measurement format

        set:
            Specifies if we want to set or retrieve (get) calibration
            measurement data. If you set this value to false (get), then
            the value of `method` will be ignored

        element:
            Specifies the element that should be measured

            Possible values:

                - Data
                - Temperature
                - Voltage
                - VSS
                - VDD
                - Internal
                - OPV

        method:
            Specifies the calibration method (Activate, Deactivate or Measure)

        dimension:
            Specifies the measurement dimension respectively the axis
            (1, 2 or 3)

        reference_voltage:
            The reference voltage in Volt

        """

        cls = type(self)

        if data:
            data_bytes = list(data[0])
            if not isinstance(data_bytes, list):
                raise ValueError(
                    "Unsupported object type for argument data: "
                    f"“{type(data_bytes)}”"
                )
            required_length = 4
            if len(data_bytes) < required_length:
                raise ValueError(
                    f"Data length of {len(data_bytes)} is too "
                    "small, at least length of "
                    f"“{required_length}” required"
                )
            self.data = data_bytes[0:4] + [0] * 4
        else:
            self.data = [0, 0, 1, 0] + [0] * 4

        if set is not None:
            method_byte = self.data[0]
            # Set get/set to 0
            method_byte &= 0b01111111
            # Set value
            method_byte |= int(set) << 7
            self.data[0] = method_byte

        if method is not None:
            if method not in cls.methods:
                raise ValueError(f"Unknown method “{method}”")
            method_byte = self.data[0]
            # Set method bits to 0
            method_byte &= 0b10011111
            # Set value
            method_byte |= cls.methods.index(method) << 5
            self.data[0] = method_byte

        if element is not None:
            if element not in cls.elements:
                raise ValueError(f"Unknown element “{element}”")
            self.data[1] = cls.elements[element]

        if dimension is not None:
            if dimension not in {1, 2, 3}:
                raise ValueError(f"Unknown dimension “{dimension}”")
            self.data[2] = dimension

        if reference_voltage is not None:
            self.data[3] = round(reference_voltage * 20) & 0xFF

    def __repr__(self) -> str:
        """
        Retrieve the textual representation of the calibration format

        Returns
        -------

        A string that describes the calibration measurement format

        Examples
        --------

        >>> CalibrationMeasurementFormat(set=False, element='Data')
        Get, Data, Dimension: 1, Reference Voltage: 0.0 V

        >>> CalibrationMeasurementFormat(set=True, method='Measure',
        ...     element='Temperature', dimension=2, reference_voltage=3.3)
        Set, Measure, Temperature, Dimension: 2, Reference Voltage: 3.3 V

        >>> CalibrationMeasurementFormat([10, 20, 30, 40])
        Get, Unknown Element, Dimension: 30, Reference Voltage: 2.0 V

        """

        cls = type(self)

        method_byte = self.data[0]

        set = method_byte >> 7
        element = cls.elements.inverse.get(self.data[1], "Unknown Element")
        dimension = self.data[2]
        reference_voltage = round(self.data[3] / 20, 1)

        parts = [
            "Set" if set else "Get",
            element,
            f"Dimension: {dimension}",
            f"Reference Voltage: {reference_voltage} V",
        ]
        if set:
            method = cls.methods[(method_byte >> 5) & 0b11]
            parts.insert(1, method)

        representation = ", ".join(parts)

        return representation


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



import unittest
import sys
import os

# Required to add peakcan
sDirName = os.path.dirname("")
sys.path.append(sDirName)
file_path = "../"
sDirName = os.path.dirname(file_path)
sys.path.append(sDirName)

import math
import time

from mytoolit.old.MyToolItCommands import (
    AdcMax,
    AdcOverSamplingRate,
    AdcReference,
    AdcVRefValuemV,
    AtvcFormat,
    byte_list_to_int,
    calcSamplingRate,
    int_to_mac_address,
    CalibMeassurementActionNr,
    CalibMeassurementTypeNr,
    MyToolItBlock,
    MyToolItConfiguration,
    MyToolItEeprom,
    MyToolItStatData,
    MyToolItStreaming,
    MyToolItTest,
    SystemCommandBlueTooth,
    SystemCommandRouting,
)
from mytoolit.old.network import Network
from mytoolit.old.MyToolItSth import TestConfig, fAdcRawDat
from mytoolit.old.MyToolItNetworkNumbers import MyToolItNetworkNr
from mytoolit.old.SthLimits import SthLimits

sVersion = TestConfig["Version"]
sLogLocation = "../../Logs/Hardware/"
sHomeLocation = "../../SimplicityStudio/v4_workspace/STH/"
sSilabsCommanderLocation = "../../SimplicityStudio/SimplicityCommander/"
sAdapterSerialNo = "440115849"
sBoardType = "BGM113A256V2"
iSensorAxis = 1
bBatteryExternalDcDc = True
uAdc2Acc = 100
iRssiMin = -75
bStuPcbOnly = True
"""
This class is used for automated internal verification of the sensory tool holder
"""


class TestSth(unittest.TestCase):
    def setUp(self):
        self.tSthLimits = SthLimits()
        SthLimits(
            iSensorAxis, bBatteryExternalDcDc, uAdc2Acc, iRssiMin, 20, 35
        )
        self.sHomeLocation = sHomeLocation
        self.sBuildLocation = sHomeLocation + "builds/" + sVersion
        self.sBootloader = (
            sHomeLocation + "builds/" + "BootloaderOtaBgm113.s37"
        )
        self.sAdapterSerialNo = sAdapterSerialNo
        self.sBoardType = sBoardType
        self.sSilabsCommander = sSilabsCommanderLocation + "commander"
        self.bError = False
        self.fileName = sLogLocation + self._testMethodName + ".txt"
        self.fileNameError = (
            sLogLocation + "Error_" + self._testMethodName + ".txt"
        )
        self.Can = Network(
            sender=MyToolItNetworkNr["SPU1"],
            receiver=MyToolItNetworkNr["STH1"],
            prescaler=self.tSthLimits.uSamplingRatePrescalerReset,
            acquisition=self.tSthLimits.uSamplingRateAcqTimeReset,
            oversampling=self.tSthLimits.uSamplingRateOverSamplesReset,
        )
        self.Can.logger.info("TestCase: " + str(self._testMethodName))
        self.vSilabsAdapterReset()
        self.Can.CanTimeStampStart(
            self._resetStu()["CanTime"]
        )  # This will also reset to STH
        self.Can.logger.info("Connect to STH")
        self.Can.bBlueToothConnectPollingName(
            MyToolItNetworkNr["STU1"], TestConfig["DevName"]
        )
        self.sStuAddr = int_to_mac_address(
            self.Can.BlueToothAddress(MyToolItNetworkNr["STU1"])
        )
        self.sSthAddr = int_to_mac_address(
            self.Can.BlueToothAddress(MyToolItNetworkNr["STH1"])
        )
        self.Can.logger.info("STU BlueTooth Address: " + self.sStuAddr)
        self.Can.logger.info("STH BlueTooth Address: " + self.sSthAddr)
        self.Can.u32EepromWriteRequestCounter(MyToolItNetworkNr["STH1"])
        iOperatingSeconds = self.Can.statisticalData(
            MyToolItNetworkNr["STU1"], MyToolItStatData["OperatingTime"]
        )[4:]
        iOperatingSeconds = byte_list_to_int(iOperatingSeconds)
        self.Can.logger.info(
            "STU Operating Seconds: " + str(iOperatingSeconds)
        )
        iOperatingSeconds = self.Can.statisticalData(
            MyToolItNetworkNr["STH1"], MyToolItStatData["OperatingTime"]
        )[4:]
        iOperatingSeconds = byte_list_to_int(iOperatingSeconds)
        self.Can.logger.info(
            "STH Operating Seconds: " + str(iOperatingSeconds)
        )
        self._statusWords()
        temp = self._SthAdcTemp()
        self.assertGreaterEqual(self.tSthLimits.iTemperatureInternalMax, temp)
        self.assertLessEqual(self.tSthLimits.iTemperatureInternalMin, temp)
        self._SthWDog()

        self.Can.logger.info(
            "_______________________________________________________________________________________________________________"
        )
        self.Can.logger.info("Start")

    def tearDown(self):
        self.Can.logger.info("Fin")
        self.Can.logger.info(
            "_______________________________________________________________________________________________________________"
        )
        if False == self.Can.bError:
            self._streamingStop()
            self._BlueToothStatistics()
            ReceiveFailCounter = self._RoutingInformation()
            iOperatingSeconds = self.Can.statisticalData(
                MyToolItNetworkNr["STU1"], MyToolItStatData["OperatingTime"]
            )[4:]
            iOperatingSeconds = byte_list_to_int(iOperatingSeconds)
            self.Can.logger.info(
                "STU Operating Seconds: " + str(iOperatingSeconds)
            )
            iOperatingSeconds = self.Can.statisticalData(
                MyToolItNetworkNr["STH1"], MyToolItStatData["OperatingTime"]
            )[4:]
            iOperatingSeconds = byte_list_to_int(iOperatingSeconds)
            self.Can.logger.info(
                "STH Operating Seconds: " + str(iOperatingSeconds)
            )
            self._statusWords()
            temp = self._SthAdcTemp()
            self.assertGreaterEqual(
                self.tSthLimits.iTemperatureInternalMax, temp
            )
            self.assertLessEqual(self.tSthLimits.iTemperatureInternalMin, temp)
            self.Can.u32EepromWriteRequestCounter(MyToolItNetworkNr["STH1"])
            self.Can.logger.info("Test Time End Time Stamp")
            self.Can.bBlueToothDisconnect(MyToolItNetworkNr["STU1"])
        else:
            ReceiveFailCounter = 0
        if 0 < ReceiveFailCounter:
            self.bError = True
        if False != self.Can.bError:
            self.bError = True
        self.Can.__exit__()
        if self._test_has_failed():
            if os.path.isfile(self.fileNameError) and os.path.isfile(
                self.fileName
            ):
                os.remove(self.fileNameError)
            if os.path.isfile(self.fileName):
                os.rename(self.fileName, self.fileNameError)

    """
    Checks if test has failed
    """

    def _test_has_failed(self):
        for _method, error in self._outcome.errors:
            if error:
                return True
        if True == self.bError:
            return True
        return False

    """
    Reset STU
    """

    def _resetStu(self, retries=5, log=True):
        self.Can.bConnected = False
        return self.Can.reset_node("STU1", retries=retries, log=log)

    """
    Reset STH
    """

    def _resetSth(self, retries=5, log=True):
        self.Can.bConnected = False
        return self.Can.reset_node("STH1", retries=retries, log=log)

    """
    Retrieve BGM113 internal Chip Temperature from the STH
    """

    def _SthAdcTemp(self):
        au8TempReturn = self.Can.calibMeasurement(
            MyToolItNetworkNr["STH1"],
            CalibMeassurementActionNr["Measure"],
            CalibMeassurementTypeNr["Temp"],
            1,
            AdcReference["1V25"],
            log=False,
        )
        iTemperature = float(byte_list_to_int(au8TempReturn[4:]))
        iTemperature /= 1000
        self.Can.logger.info("Temperature(Chip): " + str(iTemperature) + "°C")
        self.Can.calibMeasurement(
            MyToolItNetworkNr["STH1"],
            CalibMeassurementActionNr["None"],
            CalibMeassurementTypeNr["Temp"],
            1,
            AdcReference["VDD"],
            log=False,
            bReset=True,
        )
        return iTemperature

    """
    Retrieve Watch Dog Counter
    """

    def _SthWDog(self):
        WdogCounter = byte_list_to_int(
            self.Can.statisticalData(
                MyToolItNetworkNr["STH1"], MyToolItStatData["Wdog"]
            )[:4]
        )
        self.Can.logger.info("WatchDog Counter: " + str(WdogCounter))
        return WdogCounter

    """
    Get all Status Words from STH and STU
    """

    def _statusWords(self):
        self.Can.logger.info(
            "STH Status Word: {}".format(
                self.Can.node_status(MyToolItNetworkNr["STH1"])
            )
        )
        self.Can.logger.info(
            "STU Status Word: {}".format(
                self.Can.node_status(MyToolItNetworkNr["STU1"])
            )
        )

        status = self.Can.error_status(MyToolItNetworkNr["STH1"])
        if status.adc_overrun():
            self.bError = True
        self.Can.logger.info(f"STH Error Word: {status}")

        self.Can.logger.info(
            "STU Error Word: {}".format(
                self.Can.error_status(MyToolItNetworkNr["STU1"])
            )
        )

    """
    Stop any streaming
    """

    def _streamingStop(self):
        self.Can.streamingStop(
            MyToolItNetworkNr["STH1"], MyToolItStreaming["Data"]
        )
        self.Can.streamingStop(
            MyToolItNetworkNr["STH1"], MyToolItStreaming["Voltage"]
        )

    """
    Get RSSI, receive and send message counters of Bluetooth
    """

    def _BlueToothStatistics(self):
        SendCounter = self.Can.BlueToothCmd(
            MyToolItNetworkNr["STH1"], SystemCommandBlueTooth["SendCounter"]
        )
        self.Can.logger.info(
            "BlueTooth Send Counter(STH1): " + str(SendCounter)
        )
        ReceiveCounter = self.Can.BlueToothCmd(
            MyToolItNetworkNr["STH1"], SystemCommandBlueTooth["ReceiveCounter"]
        )
        self.Can.logger.info(
            "BlueTooth Receive Counter(STU1): " + str(ReceiveCounter)
        )
        Rssi = self.Can.BlueToothRssi(MyToolItNetworkNr["STH1"])
        self.Can.logger.info("BlueTooth Rssi(STH1): " + str(Rssi) + "dBm")
        SendCounter = self.Can.BlueToothCmd(
            MyToolItNetworkNr["STU1"], SystemCommandBlueTooth["SendCounter"]
        )
        self.Can.logger.info(
            "BlueTooth Send Counter(STU1): " + str(SendCounter)
        )
        ReceiveCounter = self.Can.BlueToothCmd(
            MyToolItNetworkNr["STU1"], SystemCommandBlueTooth["ReceiveCounter"]
        )
        self.Can.logger.info(
            "BlueTooth Receive Counter(STU1): " + str(ReceiveCounter)
        )
        Rssi = self.Can.BlueToothRssi(MyToolItNetworkNr["STU1"])
        self.Can.logger.info("BlueTooth Rssi(STU1): " + str(Rssi) + "dBm")

    """
    Routing information of STH send ports
    """

    def _RoutingInformationSthSend(self):
        SendCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STH1"],
            SystemCommandRouting["SendCounter"],
            MyToolItNetworkNr["STU1"],
        )
        self.Can.logger.info(
            "STH1 - Send Counter(Port STU1): " + str(SendCounter)
        )
        SendCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STH1"],
            SystemCommandRouting["SendFailCounter"],
            MyToolItNetworkNr["STU1"],
        )
        self.Can.logger.info(
            "STH1 - Send Fail Counter(Port STU1): " + str(SendCounter)
        )
        SendCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STH1"],
            SystemCommandRouting["SendLowLevelByteCounter"],
            MyToolItNetworkNr["STU1"],
        )
        self.Can.logger.info(
            "STH1 - Send Byte Counter(Port STU1): " + str(SendCounter)
        )

    """
    Routing information of STU send ports
    """

    def _RoutingInformationSthReceive(self):
        ReceiveCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STH1"],
            SystemCommandRouting["ReceiveCounter"],
            MyToolItNetworkNr["STU1"],
        )
        self.Can.logger.info(
            "STH1 - Receive Counter(Port STU1): " + str(ReceiveCounter)
        )
        ReceiveFailCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STH1"],
            SystemCommandRouting["ReceiveFailCounter"],
            MyToolItNetworkNr["STU1"],
        )
        self.Can.logger.info(
            "STH1 - Receive Fail Counter(Port STU1): "
            + str(ReceiveFailCounter)
        )
        ReceiveCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STH1"],
            SystemCommandRouting["ReceiveLowLevelByteCounter"],
            MyToolItNetworkNr["STU1"],
        )
        self.Can.logger.info(
            "STH1 - Receive Byte Counter(Port STU1): " + str(ReceiveCounter)
        )
        return ReceiveFailCounter

    """
    Routing information of STH
    """

    def _RoutingInformationSth(self):
        self._RoutingInformationSthSend()
        ReceiveFailCounter = self._RoutingInformationSthReceive()
        return ReceiveFailCounter

    """
    Routing information of STU send port SPU
    """

    def _RoutingInformationStuPortSpuSend(self):
        SendCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["SendCounter"],
            MyToolItNetworkNr["SPU1"],
        )
        self.Can.logger.info(
            "STU1 - Send Counter(Port SPU1): " + str(SendCounter)
        )
        SendCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["SendFailCounter"],
            MyToolItNetworkNr["SPU1"],
        )
        self.Can.logger.info(
            "STU1 - Send Fail Counter(Port SPU1): " + str(SendCounter)
        )
        SendCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["SendLowLevelByteCounter"],
            MyToolItNetworkNr["SPU1"],
        )
        self.Can.logger.info(
            "STU1 - Send Byte Counter(Port SPU1): " + str(SendCounter)
        )

    """
    Routing information of STU receive port SPU
    """

    def _RoutingInformationStuPortSpuReceive(self):
        ReceiveCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["ReceiveCounter"],
            MyToolItNetworkNr["SPU1"],
        )
        self.Can.logger.info(
            "STU1 - Receive Counter(Port SPU1): " + str(ReceiveCounter)
        )
        ReceiveFailCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["ReceiveFailCounter"],
            MyToolItNetworkNr["SPU1"],
        )
        self.Can.logger.info(
            "STU1 - Receive Fail Counter(Port SPU1): "
            + str(ReceiveFailCounter)
        )
        ReceiveCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["ReceiveLowLevelByteCounter"],
            MyToolItNetworkNr["SPU1"],
        )
        self.Can.logger.info(
            "STU1 - Receive Byte Counter(Port SPU1): " + str(ReceiveCounter)
        )
        return ReceiveFailCounter

    """
    Routing information of STU port SPU
    """

    def _RoutingInformationStuPortSpu(self):
        self._RoutingInformationStuPortSpuSend()
        ReceiveFailCounter = self._RoutingInformationStuPortSpuReceive()
        return ReceiveFailCounter

    """
    Routing information of STU send port STH
    """

    def _RoutingInformationStuPortSthSend(self):
        SendCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["SendCounter"],
            MyToolItNetworkNr["STH1"],
        )
        self.Can.logger.info(
            "STU1 - Send Counter(Port STH1): " + str(SendCounter)
        )
        SendCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["SendFailCounter"],
            MyToolItNetworkNr["STH1"],
        )
        self.Can.logger.info(
            "STU1 - Send Fail Counter(Port STH1): " + str(SendCounter)
        )
        SendCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["SendLowLevelByteCounter"],
            MyToolItNetworkNr["STH1"],
        )
        self.Can.logger.info(
            "STU1 - Send Byte Counter(Port STH1): " + str(SendCounter)
        )

    """
    Routing information of STU receive port STH
    """

    def _RoutingInformationStuPortSthReceive(self):
        ReceiveCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["ReceiveCounter"],
            MyToolItNetworkNr["STH1"],
        )
        self.Can.logger.info(
            "STU1 - Receive Counter(Port STH1): " + str(ReceiveCounter)
        )
        ReceiveFailCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["ReceiveFailCounter"],
            MyToolItNetworkNr["STH1"],
        )
        self.Can.logger.info(
            "STU1 - Receive Fail Counter(Port STH1): "
            + str(ReceiveFailCounter)
        )
        ReceiveCounter = self.Can.RoutingInformationCmd(
            MyToolItNetworkNr["STU1"],
            SystemCommandRouting["ReceiveLowLevelByteCounter"],
            MyToolItNetworkNr["STH1"],
        )
        self.Can.logger.info(
            "STU1 - Receive Byte Counter(Port STH1): " + str(ReceiveCounter)
        )
        return ReceiveFailCounter

    """
    Routing information of STU port STH
    """

    def _RoutingInformationStuPortSth(self):
        self._RoutingInformationStuPortSthSend()
        ReceiveFailCounter = self._RoutingInformationStuPortSthReceive()
        return ReceiveFailCounter

    """
    Routing information of system
    """

    def _RoutingInformation(self):
        ReceiveFailCounter = self._RoutingInformationSth()
        ReceiveFailCounter += self._RoutingInformationStuPortSth()
        ReceiveFailCounter += self._RoutingInformationStuPortSpu()
        return ReceiveFailCounter

    """
    Test single point in a three dimensional tube
    """

    def singleValueCompare(
        self,
        array1,
        array2,
        array3,
        middle1,
        tolerance1,
        middle2,
        tolerance2,
        middle3,
        tolerance3,
        fCbfRecalc,
    ):
        if 0 < len(array1):
            self.assertGreaterEqual(
                middle1 + tolerance1, fCbfRecalc(array1[0])
            )
            self.assertLessEqual(middle1 - tolerance1, fCbfRecalc(array1[0]))
        if 0 < len(array2):
            self.assertGreaterEqual(
                middle2 + tolerance2, fCbfRecalc(array2[0])
            )
            self.assertLessEqual(middle2 - tolerance2, fCbfRecalc(array2[0]))
        if 0 < len(array3):
            self.assertGreaterEqual(
                middle3 + tolerance3, fCbfRecalc(array3[0])
            )
            self.assertLessEqual(middle3 - tolerance3, fCbfRecalc(array3[0]))

    """
    Config ADC and determine correct sampling rate
    """

    def SamplingRate(
        self,
        prescaler,
        acquisitionTime,
        overSamplingRate,
        adcRef,
        b1=1,
        b2=0,
        b3=0,
        runTime=None,
        compare=True,
        compareRate=True,
        log=True,
        startupTime=None,
    ):
        if None == runTime:
            runTime = self.tSthLimits.uStandardTestTimeMs

        if None == startupTime:
            startupTime = self.tSthLimits.uStartupTimeMs

        Settings = self.Can.ConfigAdc(
            MyToolItNetworkNr["STH1"],
            prescaler,
            acquisitionTime,
            overSamplingRate,
            adcRef,
            log=log,
        )[1:]
        self.assertEqual(prescaler, Settings[0])
        self.assertEqual(acquisitionTime, Settings[1])
        self.assertEqual(overSamplingRate, Settings[2])
        self.assertEqual(adcRef, Settings[3])
        calcRate = calcSamplingRate(
            prescaler, acquisitionTime, overSamplingRate
        )
        if False != log:
            self.Can.logger.info("Start sending package")
        dataSets = self.Can.Can20DataSet(b1, b2, b3)
        [indexStart, indexEnd] = self.Can.streamingValueCollect(
            MyToolItNetworkNr["STH1"],
            MyToolItStreaming["Data"],
            dataSets,
            b1,
            b2,
            b3,
            runTime,
            log=log,
            StartupTimeMs=startupTime,
        )
        [array1, array2, array3] = self.Can.streamingValueArray(
            MyToolItNetworkNr["STH1"],
            MyToolItStreaming["Data"],
            dataSets,
            b1,
            b2,
            b3,
            indexStart,
            indexEnd,
        )
        self.Can.ReadThreadReset()
        samplingPoints = self.Can.samplingPoints(array1, array2, array3)
        if False != log:
            samplingPoints = self.Can.samplingPoints(array1, array2, array3)
            self.Can.logger.info("Running Time: " + str(runTime) + "ms")
            if False != startupTime:
                self.Can.logger.info(
                    "Startup Time: " + str(startupTime) + "ms"
                )
            self.Can.logger.info("Assumed Sampling Points/s: " + str(calcRate))
            samplingRateDet = 1000 * samplingPoints / (runTime)
            self.Can.logger.info(
                "Determined Sampling Points/s: " + str(samplingRateDet)
            )
            self.Can.logger.info(
                "Difference to Assumed Sampling Points: "
                + str((100 * samplingRateDet - calcRate) / calcRate)
                + "%"
            )
            self.Can.ValueLog(array1, array2, array3, fAdcRawDat, "Acc", "")
        ratM = AdcVRefValuemV[AdcReference["VDD"]] / AdcVRefValuemV[adcRef]
        ratT = 1
        if adcRef != AdcReference["VDD"]:
            ratT = self.tSthLimits.Vfs
        if False != compare:
            if 1 != ratM:
                self.Can.logger.info(
                    "Compare Ration to compensate not AVDD: " + str(ratM)
                )
            adcXMiddle = ratM * self.tSthLimits.iAdcAccXRawMiddle
            adcYMiddle = ratM * self.tSthLimits.iAdcAccYRawMiddle
            adcZMiddle = ratM * self.tSthLimits.iAdcAccZRawMiddle
            adcXTol = self.tSthLimits.iAdcAccXRawTolerance * ratT
            adcYTol = self.tSthLimits.iAdcAccYRawTolerance * ratT
            adcZTol = self.tSthLimits.iAdcAccZRawTolerance * ratT
            if 16 > AdcOverSamplingRate.inverse[overSamplingRate]:
                self.Can.logger.info(
                    "Maximum ADC Value: "
                    + str(
                        AdcMax
                        / 2
                        ** (5 - AdcOverSamplingRate.inverse[overSamplingRate])
                    )
                )
                adcXMiddle = adcXMiddle / 2 ** (
                    5 - AdcOverSamplingRate.inverse[overSamplingRate]
                )
                adcYMiddle = adcYMiddle / 2 ** (
                    5 - AdcOverSamplingRate.inverse[overSamplingRate]
                )
                adcZMiddle = adcZMiddle / 2 ** (
                    5 - AdcOverSamplingRate.inverse[overSamplingRate]
                )
            else:
                self.Can.logger.info("Maximum ADC Value: " + str(AdcMax))
            self.streamingValueCompare(
                array1,
                array2,
                array3,
                adcXMiddle,
                adcXTol,
                adcYMiddle,
                adcYTol,
                adcZMiddle,
                adcZTol,
                fAdcRawDat,
            )
        if False != compareRate:
            self.assertLess(
                runTime
                / 1000
                * calcRate
                * self.tSthLimits.uSamplingToleranceLow,
                samplingPoints,
            )
            self.assertGreater(
                runTime
                / 1000
                * calcRate
                * self.tSthLimits.uSamplingToleranceHigh,
                samplingPoints,
            )
        result = {
            "SamplingRate": calcRate,
            "Value1": array1,
            "Value2": array2,
            "Value3": array3,
        }
        return result

    """
    Turn off STH LED
    """

    def TurnOffLed(self):
        self.Can.logger.info("Turn Off LED")
        cmd = self.Can.CanCmd(
            MyToolItBlock["Configuration"], MyToolItConfiguration["Hmi"], 1, 0
        )
        message = self.Can.CanMessage20(
            cmd,
            MyToolItNetworkNr["SPU1"],
            MyToolItNetworkNr["STH1"],
            [129, 1, 2, 0, 0, 0, 0, 0],
        )
        self.Can.tWriteFrameWaitAckRetries(message)

    """
    Turn on STH LED
    """

    def TurnOnLed(self):
        self.Can.logger.info("Turn On LED")
        cmd = self.Can.CanCmd(
            MyToolItBlock["Configuration"], MyToolItConfiguration["Hmi"], 1, 0
        )
        message = self.Can.CanMessage20(
            cmd,
            MyToolItNetworkNr["SPU1"],
            MyToolItNetworkNr["STH1"],
            [129, 1, 1, 0, 0, 0, 0, 0],
        )
        self.Can.tWriteFrameWaitAckRetries(message)

    """
    Get streaming Data and collect them
    """

    def streamingTestSignalCollect(
        self,
        sender,
        receiver,
        subCmd,
        testSignal,
        testModule,
        value,
        dataSets,
        b1,
        b2,
        b3,
        testTimeMs,
        log=True,
    ):
        self.Can.logger.info("Request Test Signal: " + str(testSignal))
        self.Can.logger.info("Test Module: " + str(testModule))
        self.Can.logger.info("Test Value: " + str(testModule))
        accFormat = AtvcFormat()
        accFormat.asbyte = 0
        accFormat.b.bStreaming = 1
        accFormat.b.bNumber1 = b1
        accFormat.b.bNumber2 = b2
        accFormat.b.bNumber3 = b3
        accFormat.b.u3DataSets = dataSets
        cmd = self.Can.CanCmd(MyToolItBlock["Streaming"], subCmd, 1, 0)
        message = self.Can.CanMessage20(
            cmd, sender, receiver, [accFormat.asbyte]
        )
        if False != log:
            self.Can.logger.info("Start sending package")
        self.Can.tWriteFrameWaitAckRetries(message, retries=1)
        cmd = self.Can.CanCmd(
            MyToolItBlock["Test"], MyToolItTest["Signal"], 1, 0
        )
        message = self.Can.CanMessage20(
            cmd,
            sender,
            receiver,
            [
                testSignal,
                testModule,
                0,
                0,
                0,
                0,
                0xFF & value,
                0xFF & (value >> 8),
            ],
        )
        if False != log:
            self.Can.logger.info("Start sending test signal")
        self.Can.WriteFrame(message)
        time.sleep(0.5)
        indexStart = self.Can.GetReadArrayIndex()
        timeEnd = self.Can.get_elapsed_time() + testTimeMs
        if False != log:
            self.Can.logger.info("indexStart: " + str(indexStart))
        while self.Can.get_elapsed_time() < timeEnd:
            pass
        self.Can.streamingStop(MyToolItNetworkNr["STH1"], subCmd)
        time.sleep(0.2)  # synch to read thread
        indexEnd = (
            self.Can.GetReadArrayIndex() - 40
        )  # do not catch stop command
        if False != log:
            self.Can.logger.info("indexEnd: " + str(indexEnd))
        return [indexStart, indexEnd]

    """
    Compare collected streaming data with 3dimensional tube
    """

    def streamingValueCompare(
        self,
        array1,
        array2,
        array3,
        middle1,
        tolerance1,
        middle2,
        tolerance2,
        middle3,
        tolerance3,
        fCbfRecalc,
    ):
        samplingPoints1 = len(array1)
        samplingPoints2 = len(array2)
        samplingPoints3 = len(array3)
        samplingPoints = samplingPoints1
        if samplingPoints2 > samplingPoints:
            samplingPoints = samplingPoints2
        if samplingPoints3 > samplingPoints:
            samplingPoints = samplingPoints3
        self.Can.logger.info(
            "Received Sampling Points: " + str(samplingPoints)
        )
        self.assertGreater(samplingPoints, 0)
        for i in range(0, samplingPoints):
            if 0 != samplingPoints1:
                self.assertGreaterEqual(
                    middle1 + tolerance1, fCbfRecalc(array1[i])
                )
                self.assertLessEqual(
                    middle1 - tolerance1, fCbfRecalc(array1[i])
                )
            if 0 != samplingPoints2:
                self.assertGreaterEqual(
                    middle2 + tolerance2, fCbfRecalc(array2[i])
                )
                self.assertLessEqual(
                    middle2 - tolerance2, fCbfRecalc(array2[i])
                )
            if 0 != samplingPoints3:
                self.assertGreaterEqual(
                    middle3 + tolerance3, fCbfRecalc(array3[i])
                )
                self.assertLessEqual(
                    middle3 - tolerance3, fCbfRecalc(array3[i])
                )

    """
    Compare collected streaming data with 1dimensional tube
    """

    def streamingValueCompareSignal(self, array, testSignal):
        self.Can.logger.info(
            "Comparing to Test Signal(Test Signal/Received Signal)"
        )
        for i in range(0, len(testSignal)):
            self.Can.logger.info(
                "Point "
                + str(i)
                + ": "
                + str(testSignal[i])
                + "/"
                + str(array[i])
            )
        self.assertEqual(len(array), len(testSignal))
        for i in range(0, len(testSignal)):
            self.assertEqual(array[i], testSignal[i])

    """
    Compare calculated signal indicators of collected data e.g. Signal to noice ratio (SNR)
    """

    def siginalIndicatorCheck(
        self,
        name,
        statistic,
        quantil1,
        quantil25,
        medianL,
        medianH,
        quantil75,
        quantil99,
        variance,
        skewness,
        SNR,
    ):
        self.Can.logger.info(
            "____________________________________________________"
        )
        self.Can.logger.info("Singal Indicator Check: " + name)
        self.Can.logger.info(
            "Quantil1%, quantil25%, Median Low, Median High, Quantil75%,"
            " Quantil99%, Variance, Skewness, SNR"
        )
        self.Can.logger.info("Limit - Quantil 1%: " + str(quantil1))
        self.Can.logger.info("Limit - Quantil 25%: " + str(quantil25))
        self.Can.logger.info("Limit - Median Low: " + str(medianL))
        self.Can.logger.info("Limit - Median High: " + str(medianH))
        self.Can.logger.info("Limit - Quantil 75%: " + str(quantil75))
        self.Can.logger.info("Limit - Quantil 99%: " + str(quantil99))
        self.Can.logger.info("Limit - Variance: " + str(variance))
        self.Can.logger.info("Limit - Skewness: " + str(skewness))
        self.Can.logger.info("Limit - SNR: " + str(SNR))
        self.assertGreaterEqual(statistic["Quantil1"], quantil1)
        self.assertGreaterEqual(statistic["Quantil25"], quantil25)
        self.assertGreaterEqual(statistic["Median"], medianL)
        self.assertLessEqual(statistic["Median"], medianH)
        self.assertLessEqual(statistic["Quantil75"], quantil75)
        self.assertLessEqual(statistic["Quantil99"], quantil99)
        self.assertLessEqual(statistic["Variance"], variance)
        self.assertLessEqual(abs(statistic["Skewness"]), abs(skewness))
        SignalSNR = 20 * math.log(
            (statistic["StandardDeviation"] / AdcMax), 10
        )
        self.assertGreaterEqual(abs(SignalSNR), abs(SNR))
        self.Can.logger.info(
            "____________________________________________________"
        )

    """
    Write Page by value
    """

    def vEepromWritePage(self, iPage, value):
        au8Content = [value] * 4
        timeStamp = self.Can.get_elapsed_time()
        for offset in range(0, 256, 4):
            au8Payload = [iPage, 0xFF & offset, 4, 0] + au8Content
            self.Can.cmdSend(
                MyToolItNetworkNr["STH1"],
                MyToolItBlock["EEPROM"],
                MyToolItEeprom["Write"],
                au8Payload,
            )
        self.Can.logger.info(
            "Page Write Time: "
            + str(self.Can.get_elapsed_time() - timeStamp)
            + "ms"
        )

    """
    Read page and check content
    """

    def vEepromReadPage(self, iPage, value):
        timeStamp = self.Can.get_elapsed_time()
        for offset in range(0, 256, 4):
            au8Payload = [iPage, 0xFF & offset, 4, 0, 0, 0, 0, 0]
            index = self.Can.cmdSend(
                MyToolItNetworkNr["STH1"],
                MyToolItBlock["EEPROM"],
                MyToolItEeprom["Read"],
                au8Payload,
            )
            dataReadBack = self.Can.getReadMessageData(index)
            for dataByte in dataReadBack[4:]:
                self.assertEqual(dataByte, value)
        self.Can.logger.info(
            "Page Read Time: "
            + str(self.Can.get_elapsed_time() - timeStamp)
            + "ms"
        )

    """
    Reset the Silicion Laps Adapter
    """

    def vSilabsAdapterReset(self):
        self.Can.logger.info("Reset Adapter " + self.sAdapterSerialNo)
        sSystemCall = self.sSilabsCommander + " adapter reset "
        sSystemCall += "--serialno " + self.sAdapterSerialNo
        sSystemCall += ">>" + sLogLocation + "AdapterReset.txt"
        if os.name == "nt":
            sSystemCall = sSystemCall.replace("/", "\\")
            os.system(sSystemCall)
        # for mac and linux(here, os.name is 'posix')
        else:
            os.system(sSystemCall)
        time.sleep(4)

    """
    Duration Test of Tripple Axis Sensor: Multiple "Turn On -> Meassure -> Off"
    """

    def test0100TrippleAxisSensorOnOff(self):
        uLoopRuns = 10
        for _i in range(0, uLoopRuns):
            index = self.Can.singleValueCollect(
                MyToolItNetworkNr["STH1"], MyToolItStreaming["Data"], 1, 1, 1
            )
            [val1, val2, val3] = self.Can.singleValueArray(
                MyToolItNetworkNr["STH1"],
                MyToolItStreaming["Data"],
                1,
                1,
                1,
                index,
            )
            self.Can.ValueLog(
                val1, val2, val3, self.tSthLimits.fAcceleration, "Acc", "g"
            )

            self.singleValueCompare(
                val1,
                val2,
                val3,
                self.tSthLimits.iAdcAccXMiddle,
                self.tSthLimits.iAdcAccXTolerance,
                self.tSthLimits.iAdcAccYMiddle,
                self.tSthLimits.iAdcAccYTolerance,
                self.tSthLimits.iAdcAccZMiddle,
                self.tSthLimits.iAdcAccZTolerance,
                self.tSthLimits.fAcceleration,
            )


if __name__ == "__main__":
    sLogLocation = sys.argv[1]
    sLogFile = sys.argv[2]
    sVersion = sys.argv[3]
    if "/" != sLogLocation[-1]:
        sLogLocation += "/"
    sLogFileLocation = sLogLocation + sLogFile
    sDirName = os.path.dirname(sLogFileLocation)
    sys.path.append(sDirName)

    if not os.path.exists(sDirName):
        os.makedirs(sDirName)
    with open(sLogFileLocation, "w") as f:
        runner = unittest.TextTestRunner(f)
        unittest.main(argv=["first-arg-is-ignored"], testRunner=runner)
